
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define uint unsigned int
#define bool unsigned char
#define int64 long long int
#define DEBUG 0
#define TRUE 1
#define FALSE 0

typedef struct {int64 re,im;} gauss; gauss *GAUSS;
typedef struct {double re,im;} gaussd;
typedef struct {int64 a,b;} sqrt2; sqrt2 *SQRT2;
typedef struct {double a,b;} sqrt2d;

static bool VERBOSE=FALSE;

static void gauss_mult(gauss a,gauss b,gauss *c)
{c->re=(a.re*b.re-a.im*b.im); c->im=(a.re*b.im+b.re*a.im);}

static void gauss_sub(gauss a,gauss b,gauss *c)
{c->re=(a.re-b.re); c->im=(a.im-b.im);}

static void gauss_conj_mult(gauss a,gauss b,gauss *c)
{c->re=(a.re*b.re+a.im*b.im); c->im=(-a.re*b.im+b.re*a.im);}

static int64 gauss_norm(gauss a) {return (a.re*a.re+a.im*a.im);}

static void gauss_reduce(gauss a,gauss b,gauss *r) /* check |r| < |b| ? */
{gauss t,q; double n; gaussd af,bf,cf;
 af.re=(double) a.re; af.im=(double) a.im;
 bf.re=(double) b.re; bf.im=(double) b.im;
 cf.re=(af.re*bf.re+af.im*bf.im); cf.im=(-af.re*bf.im+bf.re*af.im);
 n=bf.re*bf.re+bf.im*bf.im; q.re=rint(cf.re/n); q.im=rint(cf.im/n);
 gauss_mult(q,b,&t); gauss_sub(a,t,r);
 if (0 && gauss_norm(*r)>gauss_norm(b))
 {printf("NOT YET REDUCED!\n"); gauss_reduce(*r,b,r);}}

static void gauss_power_mod(gauss a,int64 n,gauss b,gauss *r)
{gauss s,v; r->re=1; r->im=0; s.re=a.re; s.im=a.im;
 while (n) {if (n&1) {gauss_mult(*r,s,&v); gauss_reduce(v,b,r);}
            gauss_mult(s,s,&v); gauss_reduce(v,b,&s); n>>=1;}}

static bool gauss_is_not_square(gauss a,gauss b)
{gauss r; gauss_power_mod(a,(gauss_norm(b)-1)/2,b,&r);
 if (r.im || (r.re!=1 && r.re!=-1))
  printf("BUG: gauss_is_not_square? %lld+%lld*i %lld+%lld*i %lld+%lld*i\n",
	 a.re,a.im,b.re,b.im,r.re,r.im);
 if (r.re==1 && r.im==0) return 0; return 1;}

static int doitA_even(gauss *pf,int sz,uint64 *M) // zero padded
{int i,j,ks; uint64 e,A[sz+1]; gauss m40={4LL,0LL},r;
 if (VERBOSE) printf("Isogeny A, even d\n");
 for (i=0;i<=sz;i++) M[i]=A[i]=0;
 for (i=1;i<=sz;i++)
 {if (VERBOSE) printf("Prime %d+%di\n",pf[i].re,pf[i].im);
  gauss_reduce(pf[i],m40,&r);
  if (r.re==2 || r.im==2 || r.re==-2 || r.im==-2) M[0]|=(1ULL<<i);
  if ((gauss_norm(pf[i])&7)!=1) {M[i]|=(1ULL<<0); A[i]=1;} // j=0 term
  for (j=1;j<i;j++)
  {e=gauss_is_not_square(pf[j],pf[i]);
   A[i]+=e; A[j]+=e; M[i]|=(e<<j); M[j]|=(e<<i);
   if (pf[j].re==pf[i].re && pf[j].im==-pf[i].im)
   {e=M[i]&1; A[i]+=e; A[j]+=e; M[i]^=(e<<j); M[j]^=(e<<i);}}}
 for (i=1;i<=sz;i++) M[i]|=(A[i]&1)<<i;
 ks=dimkerF2(M,sz+1,sz+1); return ks-1;}

static int doitA_35mod8(gauss *pf,int sz,uint64 *M) // zero padded
{int i,j,ks; uint64 e,A[sz+1],s; gauss m44={4LL,4LL},r;
 if (VERBOSE) printf("Isogeny A, 3 or 5 mod 8\n");
 M[0]=1; for (i=1;i<=sz;i++) M[i]=A[i]=0;
 for (i=1;i<=sz;i++)
 {if (VERBOSE) printf("Prime %d+%di\n",pf[i].re,pf[i].im);
  gauss_reduce(pf[i],m44,&r);
  if (!((r.re==1 && r.im==0) || (r.re==-1 && r.im==0) ||
	(r.re==3 && r.im==0) || (r.re==-3 && r.im==0) ||
	(r.re==1 && r.im==2) || (r.re==1 && r.im==-2) ||
	(r.re==-1 && r.im==2) || (r.re==-1 && r.im==-2))) M[0]|=(1ULL<<i);
  s=0; if ((gauss_norm(pf[i])&7)!=1) M[i]|=(1ULL<<0);
  for (j=1;j<i;j++)
  {e=gauss_is_not_square(pf[j],pf[i]);
   A[i]+=e; A[j]+=e; M[i]|=(e<<j); M[j]|=(e<<i);
   if (pf[j].re==pf[i].re && pf[j].im==-pf[i].im)
   {e=M[i]&1; A[i]+=e; A[j]+=e; M[i]^=(e<<j); M[j]^=(e<<i);}}}
 for (i=1;i<=sz;i++) M[i]|=(A[i]&1)<<i;
 ks=dimkerF2(M,sz+1,sz+1); return ks-1;}

static int doitA_17mod8(gauss *pf,int sz,uint64 *M,uint dmod8) // 0,1 padded
{int i,j,ks; uint64 e,A[sz+2],s; gauss m44={4LL,4LL},m1n1={1LL,-1LL},r;
 if (VERBOSE) printf("Isogeny A, 1 or 7 mod 8, %d mod 8\n",dmod8);
 M[0]=1; if (dmod8==7) M[0]=3; for (i=1;i<=sz+1;i++) M[i]=A[i]=0;
 for (i=1;i<=sz;i++)
 {if (VERBOSE) printf("Prime %d+%di\n",pf[i].re,pf[i].im);
  gauss_reduce(pf[i],m44,&r);
  if (!((r.re==1 && r.im==0) || (r.re==-1 && r.im==0) ||
	(r.re==3 && r.im==0) || (r.re==-3 && r.im==0) ||
	(r.re==1 && r.im==2) || (r.re==1 && r.im==-2) ||
	(r.re==-1 && r.im==2) || (r.re==-1 && r.im==-2))) M[0]|=(1ULL<<(1+i));
  if (dmod8==1)
  {if (!((r.re==1 && r.im==0) || (r.re==-1 && r.im==0) ||
	 (r.re==0 && r.im==1) || (r.re==0 && r.im==-1) ||
	 (r.re==2 && r.im==1) || (r.re==-2 && r.im==-1) ||
	 (r.re==1 && r.im==-2) || (r.re==-1 && r.im==2))) M[1]|=(1ULL<<(1+i));}
  else
  {if (!((r.re==1 && r.im==0) || (r.re==-1 && r.im==0) ||
	 (r.re==0 && r.im==1) || (r.re==0 && r.im==-1) ||
	 (r.re==2 && r.im==-1) || (r.re==-2 && r.im==1) ||
	 (r.re==1 && r.im==2) || (r.re==-1 && r.im==-2))) M[1]|=(1ULL<<(1+i));}
  if ((gauss_norm(pf[i])&7)!=1) M[i+1]|=(1ULL<<0);
  if (gauss_is_not_square(m1n1,pf[i])) M[i+1]|=(1ULL<<1);
  for (j=1;j<i;j++)
  {e=gauss_is_not_square(pf[j],pf[i]);
   A[i]+=e; A[j]+=e; M[i+1]|=(e<<(j+1)); M[j+1]|=(e<<(i+1));
   if (pf[j].re==pf[i].re && pf[j].im==-pf[i].im)
   {e=M[i+1]&1; A[i]+=e; A[j]+=e; M[i+1]^=(e<<(j+1)); M[j+1]^=(e<<(i+1));}}}
 for (i=1;i<=sz;i++) M[i+1]|=(A[i]&1)<<(i+1);
 ks=dimkerF2(M,sz+2,sz+2); return ks-1;}

static uint gauss_factor(uint *F,int sz,gauss *pf)
{uint n=1,i;
 for (i=0;i<sz;i++)
 {if ((F[i]&3)==3) {pf[n].re=-(int64) F[i]; pf[n++].im=0;}
  else {pf[n].re=GAUSS[F[i]].re; pf[n++].im=GAUSS[F[i]].im;
        pf[n].re=GAUSS[F[i]].re; pf[n++].im=-GAUSS[F[i]].im;}} return n-1;}

/************************************************************************/

static void sqrt2_mult(sqrt2 a,sqrt2 b,sqrt2 *c)
{c->a=(a.a*b.a+2*a.b*b.b); c->b=(a.a*b.b+b.a*a.b);}

static void sqrt2_sub(sqrt2 a,sqrt2 b,sqrt2 *c)
{c->a=(a.a-b.a); c->b=(a.b-b.b);}

static void sqrt2_conj_mult(sqrt2 a,sqrt2 b,sqrt2 *c)
{c->a=(a.a*b.a-2*a.b*b.b); c->b=(-a.a*b.b+b.a*a.b);}

static int64 sqrt2_norm(sqrt2 a) {return (a.a*a.a-2*a.b*a.b);}

static void sqrt2_reduce(sqrt2 a,sqrt2 b,sqrt2 *r)
{sqrt2 q,t; double n; sqrt2d af,bf,cf;
 af.a=(double) a.a; af.b=(double) a.b;
 bf.a=(double) b.a; bf.b=(double) b.b;
 cf.a=(af.a*bf.a-2.0*af.b*bf.b); cf.b=(-af.a*bf.b+bf.a*af.b);
 n=bf.a*bf.a-2.0*bf.b*bf.b; q.a=rint(cf.a/n); q.b=rint(cf.b/n);
 sqrt2_mult(q,b,&t); sqrt2_sub(a,t,r);
 if (0) /* reduction check? */
 {int64 u=sqrt2_norm(*r),v=sqrt2_norm(b); if (u<0) u=-u; if (v<0) v=-v;
   if (u>v) {printf("NOT YET REDUCED!\n"); sqrt2_reduce(*r,b,r);}}}

static void sqrt2_power_mod(sqrt2 a,int64 n,sqrt2 b,sqrt2 *r)
{sqrt2 s,v; r->a=1; r->b=0; s.a=a.a; s.b=a.b;
 while (n) {if (n&1) {sqrt2_mult(*r,s,&v); sqrt2_reduce(v,b,r);}
            sqrt2_mult(s,s,&v); sqrt2_reduce(v,b,&s); n>>=1;}}

static bool sqrt2_is_not_square(sqrt2 a,sqrt2 b)
{sqrt2 r; sqrt2_power_mod(a,(sqrt2_norm(b)-1)/2,b,&r);
  if (r.b || (r.a!=1 && r.a!=-1))
   printf("BUG: sqrt2_is_not_square? %lld+%lld*s2 %lld+%lld*s2 %lld+%lld*s2\n",
	  a.a,a.b,b.a,b.b,r.a,r.b);
 if (r.a==1 && r.b==0) return 0; return 1;}

static int doitB_odd(sqrt2 *pf,int sz,uint64 *M,uint dmod4) // zero padded
{int i,j,ks; uint64 e,A[sz+1]; sqrt2 m40={4LL,0LL},s2={0LL,1LL},r;
 if (VERBOSE) printf("Isogeny B, odd d\n");
 for (i=0;i<=sz;i++) M[i]=A[i]=0; if (dmod4==3) M[0]=1;
 for (i=1;i<=sz;i++)
 {if (VERBOSE) printf("Prime %d+%d*s2\n",pf[i].a,pf[i].b);
  sqrt2_reduce(pf[i],m40,&r); if (r.b==2) r.b=-2; /* Hack ? */
  if (dmod4==1)
  {if (!((r.a==1 && r.b==0) || (r.a==-1 && r.b==-2) ||
	 (r.a==1 && r.b==1) || (r.a==-1 && r.b==1))) M[0]|=(1ULL<<i);}
  else
  {if (!((r.a==1 && r.b==0) || (r.a==-1 && r.b==0) ||
	 (r.a==1 && r.b==-2) || (r.a==-1 && r.b==-2))) M[0]|=(1ULL<<i);}
  if (!((r.a==1 && r.b==0) || (r.a==1 && r.b==-1) ||
	(r.a==-1 && r.b==-1) || (r.a==-1 && r.b==-2))) M[i]|=(1ULL<<0);
  for (j=1;j<=sz;j++)
  {if (i==j) continue;
   e=sqrt2_is_not_square(pf[j],pf[i]); A[i]+=e; M[i]|=(e<<j);
   if (pf[j].a==pf[i].a && pf[j].b==-pf[i].b)
   {e=sqrt2_is_not_square(s2,pf[i]); A[i]+=e; M[i]^=(e<<j);}}}
 for (i=1;i<=sz;i++) M[i]|=(A[i]&1)<<i;
 ks=dimkerF2(M,sz+1,sz+1); return ks-1;}

static int doitB_even57mod8(sqrt2 *pf,int sz,uint64 *M) // zero padded
{int i,j,ks; uint64 e,A[sz+1]; sqrt2 m40={4LL,0LL},s2={0LL,1LL},r;
 if (VERBOSE) printf("Isogeny B, 10 or 14 mod 16\n");
 for (i=0;i<=sz;i++) M[i]=A[i]=0; M[0]=1;
 for (i=1;i<=sz;i++)
 {if (VERBOSE) printf("Prime %d+%d*s2\n",pf[i].a,pf[i].b);
  sqrt2_reduce(pf[i],m40,&r); if (r.b==2) r.b=-2; /* Hack ? */
  if (!((r.a==1 && r.b==0) || (r.a==-1 && r.b==0) ||
	(r.a==1 && r.b==-2) || (r.a==-1 && r.b==-2))) M[0]|=(1ULL<<i);
  if (!((r.a==1 && r.b==0) || (r.a==1 && r.b==-1) ||
	(r.a==-1 && r.b==-1) || (r.a==-1 && r.b==-2))) M[i]|=(1ULL<<0);
  for (j=1;j<=sz;j++)
  {if (i==j) continue;
   e=sqrt2_is_not_square(pf[j],pf[i]); A[i]+=e; M[i]|=(e<<j);
   if (pf[j].a==pf[i].a && pf[j].b==-pf[i].b)
   {e=sqrt2_is_not_square(s2,pf[i]); A[i]+=e; M[i]^=(e<<j);}}}
 for (i=1;i<=sz;i++) M[i]|=(A[i]&1)<<i;
 ks=dimkerF2(M,sz+1,sz+1); return ks-1;}

static int doitB_even13mod8(sqrt2 *pf,int sz,uint64 *M,uint dmod8) // 0,1 pad
{int i,j,ks; uint64 e,A[sz+2]; sqrt2 m04={0LL,4LL},r;
 if (VERBOSE) printf("Isogeny B, 2 or 6 mod 16, %d mod 16\n",dmod8*2);
 for (i=0;i<=sz+1;i++) M[i]=A[i]=0; M[0]=1; if (dmod8==1) M[0]=3;
 for (i=1;i<=sz;i++)
 {if (VERBOSE) printf("Prime %d+%d*s2\n",pf[i].a,pf[i].b);
  sqrt2_reduce(pf[i],m04,&r); if (r.b==2) r.b=-2; /* Hack ? */
  if (!((r.a==1 && r.b==0) || (r.a==-1 && r.b==0) ||
	(r.a==3 && r.b==0) || (r.a==-3 && r.b==0) ||
	(r.a==1 && r.b==-2) || (r.a==-1 && r.b==-2) ||
	(r.a==3 && r.b==-2) || (r.a==-3 && r.b==-2))) M[0]|=(1ULL<<(i+1));
  if (dmod8==1)
  {if (!((r.a==1 && r.b==0) || (r.a==-1 && r.b==0) ||
	 (r.a==1 && r.b==1) || (r.a==-1 && r.b==-1) ||
	 (r.a==1 && r.b==-1) || (r.a==-1 && r.b==1) ||
	 (r.a==3 && r.b==-2) || (r.a==-3 && r.b==-2))) M[1]|=(1ULL<<(i+1));}
  else
  {if (!((r.a==1 && r.b==0) || (r.a==3 && r.b==0) ||
	 (r.a==1 && r.b==1) || (r.a==3 && r.b==-1) ||
	 (r.a==-1 && r.b==1) || (r.a==-3 && r.b==-1) ||
	 (r.a==3 && r.b==-2) || (r.a==1 && r.b==-2))) M[1]|=(1ULL<<(i+1));}
  if (!((r.a==1 && r.b==0) || (r.a==-3 && r.b==0) ||
	(r.a==1 && r.b==-1) || (r.a==-3 && r.b==-1) ||
	(r.a==-1 && r.b==-2) || (r.a==3 && r.b==-2) ||
	(r.a==-1 && r.b==-1) || (r.a==3 && r.b==-1))) M[i+1]|=(1ULL<<0);
  if (!((r.a==1 && r.b==0) || (r.a==3 && r.b==0) ||
	(r.a==1 && r.b==-1) || (r.a==3 && r.b==1) ||
	(r.a==-1 && r.b==-1) || (r.a==-3 && r.b==1) ||
	(r.a==1 && r.b==-2) || (r.a==3 && r.b==-2))) M[i+1]|=(1ULL<<1);
  for (j=1;j<=sz;j++)
  {if (i==j) continue;
   e=sqrt2_is_not_square(pf[j],pf[i]); A[i]+=e; M[i+1]|=(e<<(j+1));
   if (pf[j].a==pf[i].a && pf[j].b==-pf[i].b)
   {e=(M[i+1]&2)>>1; A[i]+=e; M[i+1]^=(e<<(j+1));}}}
 for (i=1;i<=sz;i++) M[i+1]|=(A[i]&1)<<(i+1);
 ks=dimkerF2(M,sz+2,sz+2); return ks-1;}

static int doitC_odd(sqrt2 *pf,int sz,uint64 *M,uint dmod4) // zero padded
{int i,j,ks; uint64 e,A[sz+1]; sqrt2 m40={4LL,0LL},s2={0LL,1LL},r;
 if (VERBOSE) printf("Isogeny C, odd d\n");
 for (i=0;i<=sz;i++) M[i]=A[i]=0; if (dmod4==3) M[0]=1;
 for (i=1;i<=sz;i++)
 {if (VERBOSE) printf("Prime %d+%d*s2\n",pf[i].a,pf[i].b);
  sqrt2_reduce(pf[i],m40,&r); if (r.b==2) r.b=-2; /* Hack ? */
  if (dmod4==1)
  {if (!((r.a==1 && r.b==0) || (r.a==-1 && r.b==0) ||
	 (r.a==1 && r.b==-2) || (r.a==-1 && r.b==-2))) M[0]|=(1ULL<<i);}
  else
  {if (!((r.a==1 && r.b==0) || (r.a==-1 && r.b==-2) ||
	 (r.a==1 && r.b==1) || (r.a==-1 && r.b==1))) M[0]|=(1ULL<<i);}
  if (!((r.a==1 && r.b==0) || (r.a==-1 && r.b==0) ||
	(r.a==1 && r.b==-2) || (r.a==-1 && r.b==-2))) {M[i]|=(1<<0); A[i]=1;}
  for (j=1;j<=sz;j++)
  {if (i==j) continue;
   e=sqrt2_is_not_square(pf[j],pf[i]); A[i]+=e; M[i]|=(e<<j);
   if (pf[j].a==pf[i].a && pf[j].b==-pf[i].b)
    {e=sqrt2_is_not_square(s2,pf[i]); A[i]+=e; M[i]^=(e<<j);}}}
 for (i=1;i<=sz;i++) M[i]|=(A[i]&1)<<i;
 ks=dimkerF2(M,sz+1,sz+1); return ks-1;}

static int doitC_even13mod8(sqrt2 *pf,int sz,uint64 *M) // zero padded
{int i,j,ks; uint64 e,A[sz+1]; sqrt2 m20={2LL,0LL},s2={0LL,1LL},r;
 if (VERBOSE) printf("Isogeny C, 2 or 6 mod 16\n");
 for (i=0;i<=sz;i++) M[i]=A[i]=0;
 for (i=1;i<=sz;i++)
 {if (VERBOSE) printf("Prime %d+%d*s2\n",pf[i].a,pf[i].b);
  sqrt2_reduce(pf[i],m20,&r);
  if (!(r.b==0 && (r.a==1 || r.a==-1))) {M[0]|=(1ULL<<i); M[i]=(1<<0); A[i]=1;}
  for (j=1;j<=sz;j++)
  {if (i==j) continue;
   e=sqrt2_is_not_square(pf[j],pf[i]); A[i]+=e; M[i]|=(e<<j);
   if (pf[j].a==pf[i].a && pf[j].b==-pf[i].b)
   {e=sqrt2_is_not_square(s2,pf[i]); A[i]+=e; M[i]^=(e<<j);}}}
 for (i=1;i<=sz;i++) M[i]|=(A[i]&1)<<i; ks=dimkerF2(M,sz+1,sz+1); return ks-1;}

static int doitC_even57mod8(sqrt2 *pf,int sz,uint64 *M,uint dmod8) // 0,1 pad
{int i,j,ks; uint64 e,A[sz+2]; sqrt2 m04={0LL,4LL},s2={0LL,1LL},r;
 if (VERBOSE) printf("Isogeny C, 10 or 14 mod 16, %d mod 16\n",dmod8*2);
 for (i=0;i<=sz+1;i++) M[i]=A[i]=0; if (dmod8==5) {M[0]=2; M[1]=1;}
 for (i=1;i<=sz;i++)
 {if (VERBOSE) printf("Prime %d+%d*s2\n",pf[i].a,pf[i].b);
  sqrt2_reduce(pf[i],m04,&r); if (r.b==2) r.b=-2; /* Hack ? */
  if (!((r.a==1 && r.b==0) || (r.a==-1 && r.b==0) ||
	(r.a==3 && r.b==0) || (r.a==-3 && r.b==0) ||
	(r.a==1 && r.b==-2) || (r.a==-1 && r.b==-2) ||
	(r.a==3 && r.b==-2) || (r.a==-3 && r.b==-2))) M[0]|=(1ULL<<(i+1));
  if (dmod8==5)
  {if (!((r.a==1 && r.b==0) || (r.a==3 && r.b==0) ||
	 (r.a==1 && r.b==1) || (r.a==3 && r.b==-1) ||
	 (r.a==-1 && r.b==1) || (r.a==-3 && r.b==-1) ||
	 (r.a==3 && r.b==-2) || (r.a==1 && r.b==-2))) M[1]|=(1ULL<<(i+1));}
  else
  {if (!((r.a==1 && r.b==0) || (r.a==-1 && r.b==0) ||
	 (r.a==1 && r.b==1) || (r.a==-1 && r.b==-1) ||
	 (r.a==1 && r.b==-1) || (r.a==-1 && r.b==1) ||
	 (r.a==3 && r.b==-2) || (r.a==-3 && r.b==-2))) M[1]|=(1ULL<<(i+1));}
  if (!((r.a==1 && r.b==0) || (r.a==-1 && r.b==0) ||
	(r.a==3 && r.b==0) || (r.a==-3 && r.b==0) ||
	(r.a==1 && r.b==-2) || (r.a==-1 && r.b==-2) ||
	(r.a==3 && r.b==-2) || (r.a==-3 && r.b==-2))) {M[i+1]|=(1<<0); A[i]=1;}
  if (!((r.a==1 && r.b==0) || (r.a==-1 && r.b==0) ||
	(r.a==1 && r.b==1) || (r.a==-1 && r.b==-1) ||
	(r.a==1 && r.b==-1) || (r.a==-1 && r.b==1) ||
	(r.a==3 && r.b==-2) || (r.a==-3 && r.b==-2))) M[i+1]|=(1ULL<<1);
  for (j=1;j<=sz;j++)
  {if (i==j) continue;
   e=sqrt2_is_not_square(pf[j],pf[i]); A[i]+=e; M[i+1]|=(e<<(j+1));
   if (pf[j].a==pf[i].a && pf[j].b==-pf[i].b)
   {e=sqrt2_is_not_square(s2,pf[i]); A[i]+=e; M[i+1]^=(e<<(j+1));}}}
 for (i=1;i<=sz;i++) M[i+1]|=(A[i]&1)<<(i+1);
 ks=dimkerF2(M,sz+2,sz+2); return ks-1;}

static uint sqrt2_factor(uint *F,int sz,sqrt2 *pf)
{uint n=1,i;
 for (i=0;i<sz;i++)
 {if ((F[i]&7)==3 || (F[i]&7)==5) {pf[n].a=(int64) F[i]; pf[n++].b=0;}
  else {pf[n].a=SQRT2[F[i]].a; pf[n++].b=SQRT2[F[i]].b;
        pf[n].a=SQRT2[F[i]].a; pf[n++].b=-SQRT2[F[i]].b;}} return n-1;}

/************************************************************************/

int KROn1[4]={0,0,0,1}, KRO2[8]={0,0,0,1,0,1,0,0},KROn2[8]={0,0,0,0,0,1,0,1};

static int kro(int A,int B)
{int r,v=0; uint k=0; while (!(B&1)) {B>>=1; v++;} if (v&1) k=KRO2[A&7]; A=A%B;
 while (1)
 {if (A==0) return(k); v=0; while (!(A&1)) {A>>=1; v++;}
  if (v&1) k^=KRO2[B&7];
  r=B-A; if (r>0) {if (A&B&2) k^=1; B=A; A=r;} else A=-r;}}

static int doit0_odd(uint *pf,uint64 *M,int n)
{int i,j; uint64 A[n],e,f,s; for (i=0;i<n;i++) M[i]=M[i+n]=A[i]=0;
 for (i=0;i<n;i++)
 {e=KRO2[pf[i]&7]; f=KROn2[pf[i]&7];
  M[i]=(e<<i)+(e<<(n+i)); M[i+n]=(e<<i)+(f<<(n+i));}
 for (i=0;i<n;i++)
 {s=0; for (j=0;j<n;j++)
  {if (i==j) continue; e=kro(pf[j],pf[i]); s+=e; A[i]|=e<<j;} A[i]|=(s&1)<<i;}
  for (i=0;i<n;i++) {M[i]^=A[i]; M[i+n]^=A[i]<<n;} return dimkerF2(M,2*n,2*n);}

static int doit0_even(uint *pf,uint64 *M,int n)
{int i,j; uint64 A[n],B[n],e,f,s;
 for (i=0;i<n;i++) M[i]=M[i+n]=A[i]=B[i]=0;
 for (i=0;i<n;i++)
 {e=KRO2[pf[i]&7]; f=KROn1[pf[i]&3];
  M[i]=(e<<i)+(e<<(n+i)); M[i+n]=(e<<i)+(f<<(n+i));}
 for (i=0;i<n;i++)
 {s=0; for (j=0;j<n;j++)
  {if (i==j) continue; e=kro(pf[j],pf[i]); s+=e; A[i]|=e<<j; B[j]|=e<<i;}
   A[i]|=(s&1)<<i; B[i]|=(s&1)<<i;}
 for (i=0;i<n;i++) {M[i]^=A[i]<<n; M[i+n]^=B[i];} return dimkerF2(M,2*n,2*n);}

/************************************************************************/

float *PVAL,*APVAL;
static void start_ap()
{float L; int p,u; int i,c; for (c=0;PRIMES[c]<100000;c++);
 PVAL=malloc(c*sizeof(float)); APVAL=malloc(c*sizeof(float));
 for (i=0;i<c;i++)
 {p=PRIMES[i]; u=2*GAUSS[p].re; if ((u&7)==6) u=-u; if ((p&7)!=1) u=-u;
  if ((p&3)==3 || p==2) u=0; L=logf(p);
  PVAL[i]=2.0*L/(1.0+p); APVAL[i]=-u*L/(1.0+p);}}

static int table8[8]={0,0,0,1,0,1,0,0};
static int kron(int A,int B)
{int r,v=0; uint k=0; if (A==0 || (B==2 && !(A&1))) return 0;
 while (!(B&1)) {B>>=1; v++;} if (v&1) k=table8[A&7]; A=A%B;
 while (1)
 {if (A==0) return 1-2*k; v=0; while (!(A&1)) {A>>=1; v++;}
  if (v&1) k^=table8[B&7]; r=B-A; if (r==0 && A!=1 && A!=-1) return 0;
  if (r>0) {if (A&B&2) k^=1; B=A; A=r;} else A=-r;}}

static float mestre_sum2(int64 s,int64 t)
{float S=0.0; int i; int64 p;
 for (i=0;PRIMES[i] && PRIMES[i]<100000;i++)
 {p=PRIMES[i]; S+=PVAL[i]+kron(((s%p)*(t%p))%p,p)*APVAL[i];} return S;}
